#############################################################################
##
#W  gapbibl.g             GAP 4 package `browse'                Thomas Breuer
##
#Y  Copyright (C)  2006,  Lehrstuhl D fuer Mathematik,  RWTH Aachen,  Germany
##


if not IsBound( HeuristicTranslationsLaTeX2XML ) then
  HeuristicTranslationsLaTeX2XML:= 0;
fi;


#############################################################################
##
#F  BrowseData.SimplifiedString( <string> )
##
##  Replace non-ASCII unicode characters by approximations
##  involving only ASCII characters.
##  (This may become unnecessary in the future.)
##
BrowseData.SimplifiedString:= function( string )
    local uni;

    uni:= Unicode( string );
    if uni = fail then
      uni:= Unicode( string, "latin1" );
    fi;
    return Encode( SimplifiedUnicodeString( uni, "ASCII" ),
                   GAPInfo.TermEncoding );
end;


#############################################################################
##
#F  BrowseBibliography( [<bibfiles>] )
##
##  <#GAPDoc Label="Bibliography_section">
##  <Section Label="sec:gapbibl">
##  <Heading>Overview of the &GAP; Bibliography</Heading>
##
##  The &GAP; documentation contains a bibliography of &GAP; related
##  publications, see&nbsp;<Cite Key="GAPBibliography"/>.
##  &Browse; provides access to this information in &GAP;,
##  using the standard facilities of the function
##  <Ref Func="NCurses.BrowseGeneric"/>, i.&nbsp;e.,
##  one can scroll in the list, search for entries, sort by year,
##  sort and categorize by authors etc.
##  <P/>
##  The &Browse; package contains a (perhaps outdated) version of this
##  bibliography.
##  One can get an updated version as follows.
##  <P/>
##  <C>wget -N http://www.gap-system.org/Doc/Bib/gap-publishednicer.bib</C>
##
##  <ManSection>
##  <Func Name="BrowseBibliography" Arg="[bibfiles]"/>
##
##  <Returns>
##  a record as returned by
##  <Ref Func="ParseBibXMLExtFiles" BookName="gapdoc"/>.
##  </Returns>
##
##  <Description>
##  This function shows the list of bibliography entries in the files
##  given by <A>bibfiles</A>, which may be a string or a list of strings
##  (denoting a filename or a list of filenames, respectively)
##  or a record (see below for the supported components).
##  <P/>
##  If no argument is given then the file <F>bibl/gap-publishednicer.bib</F>
##  in the &Browse; package directory is taken,
##  and <C>"GAP Bibliography"</C> is used as the header.
##  <P/>
##  Another perhaps interesting data file that should be available in the
##  &GAP; distribution is <F>doc/manualbib.xml</F>.
##  This file can be located as follows.
##  <P/>
##  <Example><![CDATA[
##  gap> file:= Filename( DirectoriesLibrary( "doc" ), "manualbib.xml" );;
##  ]]></Example>
##  <P/>
##  Both &BibTeX; format and the &XML; based extended format provided by the
##  &GAPDoc; package are supported by <Ref Func="BrowseBibliography"/>, see
##  Chapter&nbsp;<Ref Chap="Utilities for Bibliographies" BookName="gapdoc"/>.
##  <P/>
##  In the case of &BibTeX; format input, first a conversion to the extended
##  format takes place, via <Ref Func="StringBibAsXMLext" BookName="gapdoc"/>
##  and <Ref Func="ParseBibXMLextString" BookName="gapdoc"/>.
##  Note that syntactically incorrect entries are rejected in this conversion
##  &ndash;this is signaled with <Ref Var="InfoBibTools" BookName="gapdoc"/>
##  warnings&ndash; and that only a subset of the possible &LaTeX; markup is
##  recognized &ndash;other markup appears in the browse table except that
##  the leading backslash is removed.
##  <P/>
##  In both cases of input, the problem arises
##  that in visual mode, currently we can show only ASCII characters
##  (and the symbols in <C>NCurses.lineDraw</C>, but these are handled
##  differently, see Section&nbsp;<Ref Subsect="ssec:ncursesLines"/>).
##  Therefore, we use the function
##  <Ref Func="SimplifiedUnicodeString" BookName="gapdoc"/>
##  for replacing other unicode characters by ASCII text.
##  <P/>
##  The return value is a record as returned by
##  <Ref Func="ParseBibXMLExtFiles" BookName="gapdoc"/>,
##  its <C>entries</C> component corresponds to the bibliography entries
##  that have been <Q>clicked</Q> in visual mode.
##  This record can be used as input for
##  <Ref Func="WriteBibFile" BookName="gapdoc"/> or
##  <Ref Func="WriteBibXMLextFile" BookName="gapdoc"/>,
##  in order to produce a bibliography file,
##  or it can be used as input for
##  <Ref Func="StringBibXMLEntry" BookName="gapdoc"/>,
##  in order to produce strings from the entries, in various formats.
##  <P/>
##  The full functionality of the function
##  <Ref Func="NCurses.BrowseGeneric"/> is available.
##  <P/>
##  <Example><![CDATA[
##  gap> # sort and categorize by year, scroll down, expand a category row
##  gap> BrowseData.SetReplay( "scrrscsedddddxdddddQ" );
##  gap> BrowseBibliography();;
##  gap> # sort & categorize by authors, expand all category rows, scroll down
##  gap> BrowseData.SetReplay( "scscXseddddddQ" );
##  gap> BrowseBibliography();;
##  gap> # sort and categorize by journal, search for a journal name, expand
##  gap> BrowseData.SetReplay( Concatenation( "scrrrsc/J. Algebra",
##  >        [ NCurses.keys.ENTER ], "nxdddQ" ) );
##  gap> BrowseBibliography();;
##  gap> BrowseData.SetReplay( false );
##  ]]></Example>
##  <P/>
##  <E>Implementation remarks</E>:
##  The browse table has a dynamic header (showing the number of entries,
##  which can vary when the table is restricted), no footer and row labels;
##  one row of column labels is given by the descriptions of the table
##  columns (authors, title, year, journal).
##  <P/>
##  Row and column separators are drawn as grids
##  (cf.&nbsp;<Ref Func="NCurses.Grid"/>) composed from the special characters
##  described in Section&nbsp;<Ref Subsect="ssec:ncursesLines"/>,
##  using the component <C>work.SpecialGrid</C> of the browse table,
##  see <Ref Var="BrowseData"/>.
##  <P/>
##  For categorizing by authors, the sort parameter
##  <C>"split rows on categorizing"</C> is set to <C>"yes"</C>,
##  so the authors are distributed to different category rows,
##  hence each entry appears once for each of its authors
##  in the categorized table.
##  When a data row or an entry in a data row is selected,
##  <Q>click</Q> adds the corresponding bibliographhy entry to the result.
##  <P/>
##  The width of the title column is preset; usually titles are too long for
##  one line, and the contents of this column is formatted as a paragraph,
##  using the function <Ref Func="FormatParagraph" BookName="gapdoc"/>.
##  For the authors and journal columns, maximal widths are prescribed,
##  and <Ref Func="FormatParagraph" BookName="gapdoc"/> is used for longer
##  entries.
##  <P/>
##  For three columns, the sort parameters are defined as follows:
##  The <E>authors</E> column does not become hidden when the table is
##  categorized according to this column,
##  sorting by the <E>year</E> yields a <E>de</E>scending order,
##  and the category rows arising from these columns and the <E>journal</E>
##  column show the numbers of the data rows that belong to them.
##  <P/>
##  Those standard modes in <Ref Var="BrowseData"/> where an entry or a row
##  of the table is selected have been extended by three new actions,
##  which open a pager showing the &BibTeX;, HTML, and Text format of the
##  selected entry, respectively.
##  The corresponding user inputs are the <B>vb</B>, <B>vh</B>,
##  and <B>vt</B>.
##  <P/>
##  This function requires some of the utilities provided by the
##  &GAP; package <Package>GAPDoc</Package>
##  (see&nbsp;<Cite Key="GAPDoc"/>),
##  such as <Ref Func="FormatParagraph" BookName="gapdoc"/>,
##  <Ref Func="NormalizeNameAndKey" BookName="gapdoc"/>,
##  <Ref Func="NormalizedNameAndKey" BookName="gapdoc"/>,
##  <Ref Func="ParseBibFiles" BookName="gapdoc"/>,
##  <Ref Func="ParseBibXMLextFiles" BookName="gapdoc"/>,
##  <Ref Func="ParseBibXMLextString" BookName="gapdoc"/>,
##  <Ref Func="RecBibXMLEntry" BookName="gapdoc"/>, and
##  <Ref Func="StringBibAsXMLext" BookName="gapdoc"/>.
##  <P/>
##  The code can be found in the file <F>app/gapbibl.g</F> of the package.
##  <P/>
##  The browse table can be customized by entering a record as the argument
##  of <Ref Func="BrowseBibliography"/>,
##  with the following supported components.
##  <List>
##  <Mark><C>files</C></Mark>
##  <Item>
##    a nonempty list of filenames containing the data to be shown;
##    there is no default for this component.
##  </Item>
##  <Mark><C>filesshort</C></Mark>
##  <Item>
##    a list of the same length as the <C>files</C> component,
##    the entries are strings which are shown in the <C>"sourcefilename"</C>
##    column of the table (if this column is present);
##    the default is the list of filenames.
##  </Item>
##  <Mark><C>filecontents</C></Mark>
##  <Item>
##    a list of the same length as the <C>files</C> component,
##    the entries are strings which are shown as category values when the
##    table is categorized by the <C>"sourcefilename"</C> column;
##    the default is the list of filenames.
##  </Item>
##  <Mark><C>header</C></Mark>
##  <Item>
##    is the constant part of the header shown above the browse table,
##    the default is the first filename.
##  </Item>
##  <Mark><C>columns</C></Mark>
##  <Item>
##    is a list of records that are valid as the second argument of
##    <Ref Func="DatabaseAttributeAdd"/>,
##    where the first argument is a database id enumerator created from the
##    bibliography entries in the files in question.
##    Each entry (and also the corresponding identifier) of this database id
##    enumerator is a list of records obtained from
##    <Ref Func="ParseBibXMLextFiles" BookName="gapdoc"/> and
##    <Ref Func="RecBibXMLEntry" BookName="gapdoc"/>,
##    or from <Ref Func="ParseBibFiles" BookName="gapdoc"/>,
##    such that the list elements are regarded as equal, in the sense that
##    their fingerprints (see below) are equal.
##    The records in the <C>columns</C> list are available for constructing
##    the desired browse table, the actual appearance is controlled by the
##    <C>choice</C> component described below.
##    Columns showing authors, title, year, journal, and filename are
##    predefined and need not be listed here.
##  </Item>
##  <Mark><C>choice</C></Mark>
##  <Item>
##    a list of strings denoting the <C>identifier</C> components of those
##    columns that shall actually be shown in the table,
##    the default is <C>[ "authors", "title", "year", "journal" ]</C>.
##  </Item>
##  <Mark><C>fingerprint</C></Mark>
##  <Item>
##    a list of strings denoting component names of the entries of the
##    database id enumerator that is constructed from the data (see above);
##    two data records are regarded as equal if the values of these
##    components are equal; the default is
##    <C>[ "mrnumber", "title", "authorAsList", "editorAsList", "author" ]</C>.
##  </Item>
##  <Mark><C>sortKeyFunction</C></Mark>
##  <Item>
##    either <K>fail</K> or a function that takes a record as returned by
##    <Ref Func="RecBibXMLEntry" BookName="gapdoc"/> and returns a list
##    that is used for comparing and thus sorting the records;
##    the default is <K>fail</K>, which means that the rows of the table
##    appear in the same ordering as in the source files.
##  </Item>
##  </List>
##  </Description>
##  </ManSection>
##  </Section>
##  <#/GAPDoc>
##
BindGlobal( "BrowseBibliography", function( arg )
    local columns, choice, fingerprint, sortKeyFunction, files, filesshort,
          filecontents, header, extend_entries, entries, mapping, entities,
          file, parse, r, result, identifier, ids, idenum, entry, id, pos,
          authorswidth, l, titlewidth, journalwidth, modes, newactions,
          showaction, mode, table, ret;

    # Get the arguments.
    columns:= [];
    choice:= [ "authors", "title", "year", "journal" ];
    fingerprint:= [ "mrnumber", "title", "authorAsList", "editorAsList",
                    "author", "year", "pages" ];
    sortKeyFunction:= fail;
    if   Length( arg ) = 0 then
      files:= [ Filename( DirectoriesPackageLibrary( "browse", "bibl" ),
                          "gap-publishednicer.bib" ) ];
      filecontents:= [ "gap-publishednicer.bib" ];
      header:= "GAP Bibliography";
    elif Length( arg ) = 1 and IsString( arg[1] ) then
      files:= arg;
      filesshort:= [ BrowseData.StrippedPath( arg[1] ) ];
      filecontents:= filesshort;
      header:= files[1];
    elif Length( arg ) = 1 and IsList( arg[1] )
                           and not IsEmpty( arg[1] )
                           and ForAll( arg[1], IsString ) then
      files:= arg[1];
      filesshort:= List( files, BrowseData.StrippedPath );
      filecontents:= filesshort;
      header:= files[1];
      if 1 < Length( files ) then
        header:= Concatenation( header, ", ..." );
      fi;
    elif Length( arg ) = 1 and IsRecord( arg[1] )
                           and IsBound( arg[1].files ) then
      files:= arg[1].files;
      if not IsList( files ) or IsEmpty( files )
                             or not ForAll( files, IsString ) then
        Error( "<files> must be a nonempty list of strings" );
      fi;
      if IsBound( arg[1].filesshort ) then
        filesshort:= arg[1].filesshort;
      else
        filesshort:= List( files, BrowseData.StrippedPath );
      fi;
      if IsBound( arg[1].filecontents ) then
        filecontents:= arg[1].filecontents;
      else
        filecontents:= filesshort;
      fi;
      if IsBound( arg[1].header ) and IsString( arg[1].header ) then
        header:= arg[1].header;
      elif Length( files ) = 1 then
        header:= files[1];
      else
        header:= Concatenation( files[1], ", ..." );
      fi;
      if IsBound( arg[1].columns ) then
        columns:= ShallowCopy( arg[1].columns );
      fi;
      if IsBound( arg[1].choice ) then
        choice:= ShallowCopy( arg[1].choice );
      fi;
      if IsBound( arg[1].fingerprint ) then
        fingerprint:= ShallowCopy( arg[1].fingerprint );
      fi;
      if IsBound( arg[1].sortKeyFunction ) then
        sortKeyFunction:= arg[1].sortKeyFunction;
      fi;
    else
      Error( "usage: BrowseBibliography( [<bibfiles>] )" );
    fi;

    extend_entries:= function( entry, strings, file )
      local r;

      r:= RecBibXMLEntry( entry, "Text", strings );
      r.sourcefilename:= file;
      Add( entries, [ r, entry ] );
    end;

    # Parse the data files.
    entries:= [];
    mapping:= [];
    entities:= [];
    for file in files do
      if 3 < Length( file )
         and file{ [ Length( file )-3 .. Length( file ) ] } = ".xml" then
        # Convert the XML structure to records,
        # memorize the source.
        parse:= ParseBibXMLextFiles( file );
        if parse <> fail then
          for entry in parse.entries do
            extend_entries( entry, parse.strings, file );
          od;
          Append( mapping, parse.strings );
          Append( entities, parse.entities );
        fi;
      else
        # First convert the BibTeX format into XMLext format,
        # apply some heuristics in order to improve this,
        # and then create the records.
        parse:= ParseBibFiles( file );
        if parse <> fail then
          for entry in parse[1] do
            for r in RecNames( entry ) do
              if IsString( entry.( r ) ) then
                entry.( r ):= HeuristicTranslationsLaTeX2XML.Apply(
                                  entry.( r ) );
              fi;
            od;
            entry:= StringBibAsXMLext( entry, parse[2], parse[3]);
            if entry <> fail then
              entry:= ParseBibXMLextString( entry ).entries[1];
              extend_entries( entry, parse[2], file );
            fi;
          od;
          Append( mapping, TransposedMat( parse{ [ 2, 3 ] } ) );
        fi;
      fi;
    od;

    result:= rec( entries:= [],
                  strings:= [],
                  entities:= [], );
    if IsEmpty( entries ) then
      return result;
    fi;

    # Sort the entries if required.
    if sortKeyFunction <> fail then
      SortParallel( List( entries, e -> sortKeyFunction( e[1] ) ), entries );
    fi;

    # Construct a database id enumerator from the contents.
    # (Deal with duplicates by collecting entries with the same identifier.)
    identifier:= function( entry )
      local id, fld;

      id:= rec();
      for fld in fingerprint do
        if IsBound( entry.( fld ) ) then
          id.( fld ):= entry.( fld );
        fi;
      od;
      return id;
    end;

    ids:= [];
    idenum:= [];
    for entry in entries do
      id:= identifier( entry[1] );
      pos:= Position( ids, id );
      if pos = fail then
        Add( ids, id );
        Add( idenum, [ entry ] );
      else
        Add( idenum[ pos ], entry );
      fi;
    od;
    idenum:= DatabaseIdEnumerator( rec(
        identifiers:= idenum,
        entry:= function( dbidenum, id ) return id; end,
      ) );

    # Construct database attributes for the columns to be displayed.
    authorswidth:= 35;
    DatabaseAttributeAdd( idenum, rec(
      identifier:= "authors",
      type:= "values",
      create:= function( attr, id )
        local normalize, auth;

        normalize:= true;
        id:= id[1][1];
        if   IsBound( id.authorAsList ) then
          auth:= id.authorAsList;
        elif IsBound( id.editorAsList ) then
          auth:= id.editorAsList;
        elif IsBound( id.author ) then
          auth:= id.author;
        elif IsBound( id.organization ) then
          auth:= id.organization;
          normalize:= false;
        elif IsBound( id.editor ) then
          auth:= id.editor;
        elif IsBound( id.publisher ) then
          auth:= id.publisher;
          normalize:= false;
        else
          auth:= "(no author, organization, editor, publisher)";
          normalize:= false;
        fi;
        if normalize and IsString( auth ) then
          # The value comes from a `.bib' file;
          # the value obtained from an `.xml' file need not be changed.
          auth:= NormalizedNameAndKey( auth )[4];
        fi;
        if IsString( auth ) then
          auth:= [ BrowseData.SimplifiedString( auth ) ];
        else
          auth:= List( auth, y -> BrowseData.SimplifiedString(
                                      Concatenation( y[1], ", ", y[2] ) ) );
        fi;
        auth:= List( auth, NormalizedWhitespace );
        if authorswidth < MaximumList( List( auth, Length ) ) then
          auth:= Concatenation( List( auth, l -> SplitString( FormatParagraph(
            BrowseData.SimplifiedString( l ), authorswidth, "left" ),
            "\n" ) ) );
        fi;
        return auth;
      end,
      viewValue:= x -> rec( rows:= x, align:= "tl" ),
      align:= "l",
      widthCol:= authorswidth,
      sortParameters:= [ "hide on categorizing", "no",
                         "add counter on categorizing", "yes",
                         "split rows on categorizing", "yes" ],
    ) );

    titlewidth:= 40;
    DatabaseAttributeAdd( idenum, rec(
      identifier:= "title",
      type:= "values",
      create:= function( attr, id )
        if   IsBound( id[1][1].title ) then
          return id[1][1].title;
        elif IsBound( id[1][1].booktitle ) then
          return id[1][1].booktitle;
        else
          return "(no title or booktitle)";
        fi;
        end,
      viewValue:= x -> rec(
              rows:= SplitString( ReplacedString( FormatParagraph(
                       BrowseData.SimplifiedString( x ),
                                titlewidth, "left" ), "\n", " \n" ), "\n" ),
              align:= "tl" ),
      align:= "l",
      widthCol:= titlewidth,
    ) );

    DatabaseAttributeAdd( idenum, rec(
      identifier:= "year",
      type:= "values",
      create:= function( attr, id )
        if IsBound( id[1][1].year ) then
          return id[1][1].year;
        else
          return "(no year)";
        fi;
        end,
      viewValue:= x -> rec( rows:= [ x ], align:= "tr" ),
      align:= "r",
      sortParameters:= [ "direction", "descending",
                         "add counter on categorizing", "yes" ],
    ) );

    journalwidth:= 35;
    DatabaseAttributeAdd( idenum, rec(
      identifier:= "journal",
      type:= "values",
      create:= function( attr, id )
        local jour;

        if IsBound( id[1][1].journal ) then
          jour:= [ NormalizedWhitespace( BrowseData.SimplifiedString(
                       id[1][1].journal ) ) ];
          if journalwidth < Length( jour[1] ) then
            jour:= SplitString( ReplacedString( FormatParagraph( jour[1],
                                  journalwidth, "left" ), "\n", " \n" ),
                     "\n" );
          fi;
          return jour;
        else
          return [ "(no journal)" ];
        fi;
        end,
      viewValue:= x -> rec( rows:= x, align:= "tl" ),
      align:= "l",
      widthCol:= journalwidth,
      sortParameters:= [ "add counter on categorizing", "yes" ],
    ) );

    DatabaseAttributeAdd( idenum, rec(
      identifier:= "sourcefilename",
      viewLabel:= "filename",
      type:= "values",
      create:= function( attr, id )
        local rows, r;

        rows:= [];
        for r in id do
          if IsBound( r[1].sourcefilename ) and
             not r[1].sourcefilename in rows then
            Add( rows, r[1].sourcefilename );
          fi;
        od;
        return rows;
        end,
      viewValue:= value -> rec( rows:= List( value,
          x -> BrowseData.ReplacedEntry( x, files, filesshort ) ),
                                align:= "tl" ),
      categoryValue:= value -> List( value,
          x -> BrowseData.ReplacedEntry( x, files, filecontents ) ),
      align:= "l",
      sortParameters:= [ "hide on categorizing", "no",
                         "add counter on categorizing", "yes",
                         "split rows on categorizing", "yes" ],
    ) );

    # Add user defined columns.
    for l in columns do
      DatabaseAttributeAdd( idenum, l );
    od;

    # Construct the extended modes if necessary.
    if not IsBound( BrowseData.defaults.work.customizedModes.gapbibl ) then
      # Create a shallow copy of each default mode for `Browse', and add
      # new actions to those modes where a row or an entry is selected:
      # - vb: Show BibTeX format of the selected entry in a pager
      # - vh: Show HTML format of the selected entry in a pager
      # - vt: Show text format of the selected entry in a pager
      modes:= List( BrowseData.defaults.work.availableModes,
                    BrowseData.ShallowCopyMode );
      BrowseData.defaults.work.customizedModes.gapbibl:= modes;
      newactions:= [ [ "vb", "BibTeX" ],
                     [ "vh", "HTML" ],
                     [ "vt", "Text" ] ];
      showaction:= pair -> [ [ pair[1] ],
        rec( helplines:= [ Concatenation( "show ", pair[2],
                             " format of the selected entry in a pager" ) ],
             action:= function( t )
               local pos, disp;

               if t.dynamic.selectedEntry <> [ 0, 0 ] then
                 pos:= t.dynamic.indexRow[ t.dynamic.selectedEntry[1] ] / 2;
                 disp:= StringBibXMLEntry( t.work.entries[ pos ][1][2],
                                           pair[2], t.work.mapping );
                 disp:= SimplifiedUnicodeString( Unicode( disp ), "ASCII" );
                 disp:= Encode( disp, GAPInfo.TermEncoding );
                 NCurses.hide_panel( t.dynamic.statuspanel );
                 NCurses.Pager( disp );
                 NCurses.show_panel( t.dynamic.statuspanel );
               fi;
               t.dynamic.changed:= true;
             end ) ];
      newactions:= List( newactions, showaction );
      for mode in modes do
        if mode.name in [ "select_row", "select_entry",
                          "select_row_and_entry",
                          "select_column_and_entry" ] then
          BrowseData.SetActions( mode, newactions );
        fi;
      od;
    else
      modes:= BrowseData.defaults.work.customizedModes.gapbibl;
    fi;

    # Construct the browse table.
    table:= BrowseTableFromDatabaseIdEnumerator( idenum, [], choice,
                t -> BrowseData.HeaderWithRowCounter( t, header,
                          Length( idenum.identifiers ) ) );
    table.work.cacheEntries:= true;
    table.work.Click:= rec(
          select_entry:= rec(
            helplines:= [ "add the BibTeX or XML entry to the result list" ],
            action:= function( t )
            if t.dynamic.selectedEntry <> [ 0, 0 ] then
              Add( t.dynamic.Return,
                   t.dynamic.indexRow[ t.dynamic.selectedEntry[1] ] / 2 );
            fi;
          end ),
          select_row:= ~.select_entry,
        );
    table.work.availableModes:= modes;
    table.dynamic.Return:= [];
    table.dynamic.activeModes:= [ First( modes, x -> x.name = "browse" ) ];

    table.work.entries:= idenum.identifiers;
    table.work.mapping:= mapping;

    # Show the table.
    ret:= DuplicateFreeList( NCurses.BrowseGeneric( table ) );

    # Construct the return value.
    if not IsEmpty( ret ) then
      result.strings:= mapping;
      result.entities:= entities;
      result.entries:= List( idenum.identifiers{ ret }, x -> x[1][2] );
    fi;

    return result;
end );


if HeuristicTranslationsLaTeX2XML = 0 then
  Unbind( HeuristicTranslationsLaTeX2XML );
fi;


#############################################################################
##
##  Add the Browse application to the list shown by `BrowseGapData'.
##
BrowseGapDataAdd( "GAP Bibliography", BrowseBibliography, true, "\
an overview of GAP related publications, \
in a browse table whose columns show authors, title, year, and journal; \
the return value is a record encoding the bibliography entries \
of the clicked packages" );


#############################################################################
##
#E

